# -*- coding: utf-8 -*-
"""
Created on Tue Aug 30 14:30:41 2016

@author: akibdpt
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import super
"""
from future import standard_library
standard_library.install_aliases()
"""
from .multiplyForm import multiplyForm
from ..transform.graphTransform import graphTransform, graphTransform3D
"""
 * A meta-form denotes a form of a {DisjunctiveSort}.
 * It is composed of zero, one or more forms corresponding to the component
 * sorts of the disjunctive sort. The operational behavior of a meta-form is
 * defined by the individual behaviors of the element forms.
 * A meta-form is maximal if it contains at most one form
 * corresponding to each component sort and each element form is maximal.
 * 
 * The meta-form class extends on the {MultiplyForm} class and
 * specifies the operational behavior of meta-forms in terms of the behaviors
 * of its constituent element forms.
"""
class metaForm(multiplyForm):
    type = 'form-metaForm'
    
    def __init__(self, sort, *args, **kwargs):
        super(metaForm, self).__init__(sort)
        if sort.type == 'sort-compound':
            self.compound = True
        else:
            self.compound = False
    
    """
    def __str__(self):
        s = '{'
        printComma = False
        for i in self.elements:
            if printComma:
                s += ', '
            else:
                printComma = True
            s += i.sort.name+': '+str(i)
        s += '}'
        return s
    """
        
    def toString(self, string = '', printKey = False):
        string += '  '
        s = '{\n' + string
        printComma = False
        for i in self.elements:
            if printComma:
                s += ',\n' + string
            else:
                printComma = True
            s += i.sort.name+': '+i.toString(string, printKey)
        s += '\n'+ string[:-2]+ '}'
        return s
        
    def totalLength(self):
        length = 0
        for i in self:
            length += i.totalLength()
        return length
        
    def oneLineString(self, printKey = False):
        s = '{'
        printComma = False
        for i in self.elements:
            if printComma:
                s += ', '
            else:
                printComma = True
            s += i.sort.name+': '+ i.oneLineString(printKey)
        s += '}'
        return s
        

    def __eq__(self, other):
        if other is None or self.type != other.type or self.sort != other.sort or len(self.elements) != len(other.elements):
            return False
        
        if not self.maximal:
            self.maximalize()
        if not other.maximal:
            other.maximalize()
            
        lenSelf = 0
        for i in self:
            lenSelf += len(i)
        lenOther = 0
        for i in other:
            lenOther += len(i)
        if lenSelf != lenOther:
            return False
            
        for i in self.elements:
            check = False
            for j in other.elements:
                if i.type != j.type or i.sort != j.sort:
                    continue
                if i == j:
                    check = True
                    break
            if check == False:
                return False
        return True
        
    
    """
    print out the form for easier assessment
    structure:
      |disjunctive: r elements
        |form 1 sort name: n elements
         _ individual 1
         _ individual 2
             ...
         _ individual n
          |attribute1 sort name: m elements
           _ attribute1 1
               ...
          |attribute2 sort name: 1 elements
           _ attribute2
        |form 2 sort name: p elements
         _ individual 1
         _ individual 2
             ...
         _ individual p
        ...
        |form ??? sort name: ??? elements
            ...
    """
    def printE(self, string = ''):
        print(string + '|' + self.sort.name+':',len(self.elements),'elements')
        for ele in self.elements:
            ele.printE(string+'  ')
        return
        
        """
    def printRhino(self, dX = 0, dY = 0):
        for ele in self.elements:
            if ele.type in ('form-metaForm','form-interval','form-discrete','form-coForm','form-areal','form-cyclical','form-curve'):
                ele.printRhino(dX, dY)
        """

    def rhinoString(self, dX = 0, dY = 0):
        s = ''
        for ele in self.elements:
            if ele.type in ('form-metaForm','form-interval','form-discrete','form-coForm','form-areal','form-cyclical','form-curve'):
                s += ele.rhinoString(dX, dY)
        return s
        

    def accept(self, visitor, direction=-1):
        if not self.compound and isinstance(direction, int) and -1 == direction:
            direction = self.sort.name
    
        if visitor.visitEnter(self, direction) and not self.isEmpty():
            for element in self.elements:
                element.accept(visitor, direction)
        visitor.visitLeave(self)
        return

    def splitElements(self):
        splits = []
        for i in self.elements:
            splits += i.splitElements()
        return splits
        
    def switchSort(self, newSort):
        if not newSort.contains(self.sort):
            raise ValueError(self.sort.name+' is not a subset of '+newSort.name)
        form = newSort.newForm()
        for i in self.elements:
            form.elements.append(i.duplicate())
        return form
        
    def splitSort(self, newSort, complementSort = None):
        if not self.sort.contains(newSort):
            raise ValueError(newSort.name+' is not a subset of '+self.sort.name)
        form = newSort.newForm()
        
        if complementSort is None:
            complementSort = self.sort.complementSort(newSort)
        else:
            for i in self.sort.component:
                if i not in newSort.component and i not in complementSort.component:
                    raise ValueError('Sorts '+newSort.name+' and '+complementSort.name+' do not sum up to become '+self.sort.name)
        complementForm = complementSort.newForm()
        for i in self.elements:
            if i.sort in newSort.component:
                form.elements.append(i.duplicate())
            else:
                complementForm.elements.append(i.duplicate())
        return form, complementForm
                
    def mergeSort(self, otherForm, newSort):
        if not newSort.contains(self.sort):
            raise ValueError(self.sort.name+' is not a subset of '+newSort.name)
        if not newSort.contains(otherForm.sort):
            raise ValueError(self.sort.name+' is not a subset of '+newSort.name)
        form = newSort.newForm()
        for i in self:
            form.insert(i)
        for i in otherForm:
            form.insert(i)
        return form
                
    """
    inserting an element into the form
    """
    def insert(self, element):
        if self.compound:
            if element.type == 'form-metaForm':
                if not self.sort.contains(element.sort):
                    print('~~~~~~~')
                    print("ERROR: Argument sorts do not adhere to containment relationship:")
                    print(self.sort.name,"does not contain",element.sort.name)
                    return
                for i in self.elements:
                    if i.sort == element.sort:
                        for j in element:
                            i.insert(j)
                        return
                self.elements.append(element)
                return
            for i in self.elements:
                if i.sort.contains(element.sort):
                    i.insert(element)
                    return
            for i in self.sort.component:
                if i.contains(element.sort):
                    newForm = metaForm(i)
                    newForm.insert(element)
                    self.elements.append(newForm)
                    return
            return
        if 'form' in element.type:
            if not self.sort.contains(element.sort):
                print('~~~~~~~')
                print("ERROR: Argument sorts do not adhere to containment relationship:")
                print(self.sort.name,"does not contain",element.sort.name)
                return
            self.maximal = False
            if element.type == 'form-metaForm':
                for i in element:
                    self.insert(i)
                return
            index = 0
            for i in self.elements:
                if i.type == 'form-metaForm':
                    if i.sort.contains(element.sort):
                        i.insert(element)
                        return
                elif i.sort.baseSort().dimension() == element.sort.baseSort().dimension():
                    if i.sort.name == element.sort.name:
                        i.merge(element)
                        return
                    elif element.sort.name < i.sort.name:
                        self.add(index, element)
                        return
                elif i.sort.baseSort().dimension() > element.sort.baseSort().dimension():
                    self.add(index, element)
                    return
                index = index + 1
            self.add(index, element)
        else:
            if not self.sort.contains(element.sort):
                print('!!!!!!!!!!')
                print(self.sort.name, element.sort.name)
                print("Argument sorts do not adhere to containment relationship")
                return
            self.maximal = False
            index = 0
            for i in self.elements:
                if i.sort.baseSort().dimension() == element.sort.baseSort().dimension():
                    if i.sort.name == element.sort.name:
                        i.insert(element)
                        return
                    elif element.sort.name < i.sort.name:
                        form = element.sort.newForm()
                        form.insert(element)
                        self.add(index, form)
                        return
                elif i.sort.baseSort().dimension() > element.sort.baseSort().dimension():
                    form = element.sort.newForm()
                    form.insert(element)
                    self.add(index, form)
                    return
                index = index + 1
            form = element.sort.newForm()
            form.insert(element)
            self.add(index, form)
        return
               
    """
    maximalize the form:
    _remove Nil elements
    _reorder the elements in the form in ascending order
    """
    
    def maximalize(self):
        if self.maximal:
            return
        self.maximal = True
    
        for i in self.elements:
            i.maximalize()
        i = self.first()
        disorganized = []
        index = 1
        j = self.get(index)
        while j is not None:
            if i.sort.baseSort().dimension() == j.sort.baseSort().dimension():
                if j.sort.name == i.sort.name:
                    i.sum(j)
                    self.remove(j)
                    j = self.get(index)
                elif i.sort.name > j.sort.name:
                    self.remove(j)
                    disorganized.append(j)
                    j = self.get(index)
                else:
                    index += 1
                    i = j
                    j = self.get(index)
            elif i.sort.baseSort().dimension() > j.sort.baseSort().dimension():
                self.remove(j)
                disorganized.append(j)
                j = self.get(index)
            else:
                index += 1
                i = j
                j = self.get(index)
        for i in disorganized:
            self.insert(i)
        return
        
    """
    sum two forms (OR)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.sum(formB)
      -->formA: a, b, c, d
    """
    def sum(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            if self.sort.contains(other.sort):
                for i in self.elements:
                    if i.sort == other.sort or i.sort.contains(other.sort):
                        i.sum(other)
                return
            raise ValueError('not of the same sort')
        
        self.maximalize()
        other.maximalize()
        
        thisIndex = 0
        otherIndex = 0
        thisForm = self.first()
        otherForm = other.first()
        while thisForm is not None and otherForm is not None:   
            if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                if thisForm.sort.name == otherForm.sort.name:
                    #if has attribute, sum
                    thisForm.sum(otherForm)
                    thisIndex += 1
                    otherIndex += 1
                    thisForm = self.get(thisIndex)                            
                    otherForm = other.get(otherIndex)
                elif thisForm.sort.name > otherForm.sort.name:
                    self.add(thisIndex, otherForm)      
                    otherIndex += 1                      
                    otherForm = other.get(otherIndex)
                else:
                    thisIndex += 1
                    thisForm = self.get(thisIndex)                            
            elif thisForm.sort.baseSort().dimension() > otherForm.sort.baseSort().dimension():
                self.add(thisIndex, otherForm)
                thisIndex += 1                 
                otherIndex += 1
                otherForm = other.get(otherIndex)
            else:
                thisIndex += 1
                thisForm = self.get(thisIndex)                            
                 
        while otherForm is not None:
            self.elements.append(otherForm)                            
            otherIndex += 1
            otherForm = other.get(otherIndex)
        return
        
    """
    difference of two forms
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.difference(formB)
      -->formA: a
    """
    def difference(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if self.sort != other.sort:
            if self.sort.contains(other.sort):
                for i in self.elements:
                    if i.sort == other.sort or i.sort.contains(other.sort):
                        i.difference(other)
                return
            raise ValueError('not of the same sort')
        
        self.maximalize()
        other.maximalize()
            
        thisIndex = 0
        otherIndex = 0
        thisForm = self.first()
        otherForm = other.first()
        while thisForm is not None and otherForm is not None:
            if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                if thisForm.sort.name == otherForm.sort.name:
                    #if has attribute, calculate difference
                    thisForm.difference(otherForm)
                    if thisForm.isEmpty():
                        self.pop(thisIndex)
                        thisForm = self.get(thisIndex)
                    otherIndex += 1
                    otherForm = other.get(otherIndex)
                elif thisForm.sort.name > otherForm.sort.name:                            
                    otherIndex += 1
                    otherForm = other.get(otherIndex)
                else:                             
                    thisIndex += 1
                    thisForm = self.get(thisIndex)
            elif thisForm.sort.baseSort().dimension() > otherForm.sort.baseSort().dimension():
                otherIndex += 1
                otherForm = other.get(otherIndex)
            else:
                thisIndex += 1
                thisForm = self.get(thisIndex)
        return        
        
    """
    product of two forms (AND)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.product(formB)
      -->formA:    b, c
    """
    def product(self, other):
        if self is other:
            return
        if not self.sort.__eq__(other.sort):
            if self.sort.contains(other.sort):
                for i in self.elements:
                    if i.sort == other.sort or i.sort.contains(other.sort):
                        i.product(other)
                return
            raise ValueError('not of the same sort')
        differenceForm = self.duplicate()
        differenceForm.difference(other)
        self.difference(differenceForm)
        return

        self.maximalize()
        other.maximalize()
                       
        thisIndex = 0
        otherIndex = 0
        thisForm = self.first()
        otherForm = other.first()
        while thisForm is not None and otherForm is not None:
            if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                if thisForm.sort.name == otherForm.sort.name:
                    #if has attribute, product
                    thisForm.product(otherForm)
                    if thisForm.isEmpty():
                        self.pop(thisIndex)
                        thisForm = self.get(thisIndex)
                        otherIndex += 1
                        otherForm = other.get(otherIndex)
                    else:
                        thisIndex += 1
                        otherIndex += 1
                        thisForm = self.get(thisIndex)
                        otherForm = other.get(otherIndex)
                elif thisForm.sort.name > otherForm.sort.name:                
                    otherIndex += 1
                    otherForm = other.get(otherIndex)
                else:                             
                    self.pop(thisIndex)
                    thisForm = self.get(thisIndex)
            elif thisForm.sort.baseSort().dimension() > otherForm.sort.baseSort().dimension():
                otherIndex += 1
                otherForm = other.get(otherIndex)
            else:
                self.pop(thisIndex)
                thisForm = self.get(thisIndex)
        while thisForm is not None:   
            self.pop(thisIndex)
            thisForm = self.get(thisIndex)
        return

        
    """
    check if this form is a part of other form
    """
    def partOf(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort == other.sort:
            if other.sort.contains(self.sort):
                for i in other:
                    if i.sort == self.sort and self.partOf(i):
                        return True
                return False
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
                
        for thisForm in self.elements:
            if thisForm.isEmpty():
                continue
            check = False
            for otherForm in other.elements:
                if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                    if thisForm.sort.name == otherForm.sort.name:
                        if thisForm.partOf(otherForm):
                            check = True
            if check == False:
                return False
        return True
        
    """
    editing the form in a certain procedure(editor)
    mainly use for transforming the form elements with transformation matrix(es)
    """        
    def edit(self, editor):
        prev = None
        if not self.compound and self.assoc is None:
            if editor.current is not None:
                prev = editor.current
            editor.current = self.sort.name

        self.maximal = False
        result = self.sort.newForm()
        for i in self.elements:
            result.insert(i.edit(editor))
        """
        if not self.compound and editor.type == 'transformSet':
            transf = editor.retrieveTransform(graphTransform)
            if transf is not None:
                for i in transf.extra:
                    result.insert(i)
            transf = editor.retrieveTransform(graphTransform3D)
            if transf is not None:
                for i in transf.extra:
                    result.insert(i)
        """
        if prev is not None:
            editor.current = prev
        return result

    """
    check if this form is contained by the compareForm after transforming
    with transformSet
    """
    def checkTransform(self, transformSet, compareForm):
        prev = None
        if not self.compound and self.assoc is None:
            if transformSet.current is not None:
                prev = transformSet.current
            transformSet.current = self.sort.name
            
        if compareForm.compound:
            if self.compound:
                for i in self.elements:
                    if i.isEmpty():
                        continue
                    check = False
                    for j in compareForm.elements:
                        if i.sort.name == j.sort.name:
                            if not i.checkTransform(transformSet, j):
                                return False
                            check = True
                            break
                    if not check:
                        return False
                if prev is not None:
                    transformSet.current = prev
                return True
            else:
                for j in compareForm.elements:
                    if self.sort == j.sort:
                        if not self.checkTransform(transformSet, j):
                            return False
                        return True
                return False
        for i in self.elements:
            if i.isEmpty():
                continue
            if not i.checkTransform(transformSet, compareForm):
                return False
        if prev is not None:
            transformSet.current = prev
        return True      
        
    def checkPartOf(self, other, flagDesc = True):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort == other.sort:
            if other.sort.contains(self.sort):
                for i in other:
                    if i.sort == self.sort and self.partOf(i):
                        return True
                return False
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        self.maximalize()
        other.maximalize()
                
        for thisForm in self.elements:
            if thisForm.isEmpty():
                continue
            check = False
            for otherForm in other.elements:
                if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                    if thisForm.sort.name == otherForm.sort.name:
                        if thisForm.checkPartOf(otherForm, flagDesc):
                            check = True
                        
            if check == False:
                return False
        return True
        

