# -*- coding: utf-8 -*-
"""
Created on Wed Jan 11 15:22:42 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from .tree import tree
from .symbols import symbols
from ...io.parseReader import parseReader, ParsingError

from .tuple import tupleDesc
from .kleeneTuple import kleeneTuple

class tupleExpression(tree):
    def __init__(self, left, right):
        self.type = 'description-tupleExpression'
        self.expr = []
        self.bracket = symbols.NONE
        self.endBracket = symbols.NONE
        self.operator = symbols.PLUS
        if right is None:
            raise TypeError('right entity cannot be None')
        self.expr.append(left)
        self.expr.append(right)
        
    @staticmethod
    def _initBracket(bracket, entity):
        tplExpr = tupleExpression(entity, '')
        tplExpr.expr.pop(-1)
        tplExpr.bracket = bracket
        if bracket == symbols.ANGLE_OPEN:
            tplExpr.endBracket = symbols.ANGLE_CLOSE
        elif bracket == symbols.SQUARE_OPEN:
            tplExpr.endBracket = symbols.SQUARE_CLOSE
        elif bracket == symbols.SET_OPEN:
            tplExpr.endBracket = symbols.SET_CLOSE
        elif bracket == symbols.PAREN_OPEN:
            tplExpr.endBracket = symbols.PAREN_CLOSE
        tplExpr.operator = symbols.NONE
        return tplExpr
        
    @staticmethod
    def _initOperator(operator):
        tplExpr = tupleExpression('', '')
        del tplExpr.expr[:]
        tplExpr.bracket = symbols.NONE
        tplExpr.endBracket = symbols.NONE
        tplExpr.operator = operator
        return tplExpr
    
    def left(self):
        return self.expr[0]
        
    def right(self):
        return self.expr[1]
    
    def __eq__(self, other):
        if not other.type == 'description-tupleExpression':
            return False
        if self.operator != other.operator:
            return False
        if len(self.expr) != len(other.expr):
            return False
        for i in range(len(self.expr)):
            if self.expr[i] != other.expr[i]:
                return False
        return True
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1

    def compareValue(self, other):
        if other.type == 'description-tupleExpression':
            if self.left() is None and other.left() is not None:
                return -1
            if self.left() is not None:
                if other.left() is None:
                    return 1
                c = self.left().compareValue(other.left())
                if c != 0:
                    return c
            if self.operator > other.operator:
                return 1
            if self.operator < other.operator:
                return -1
            if self.operator == symbols.PLUS:
                c = self.right().compareValue(other.right())
            else:
                for i in range(len(self.expr)):
                    if i>= len(other.expr):
                        break
                    if self.expr[i] > other.expr[i]:
                        return 1
                    elif self.expr[i] < other.expr[i]:
                        return -1
                c = len(self.expr) - len(other.expr)
        elif (other.type == 'description-tuple' or other.type == 'description-function' or
              other.type == 'description-reference' or other.type == 'description-parameter'):
            if self.left() is None:
                return -1
            c = self.left().compareValue(other)
            if c == 0:
                c = -1
        else:
            c = hash(self.__class__) - hash(other.__class__)
        return (c>0)-(c<0)

    def __str__(self):
        s = ''
        if self.bracket != symbols.NONE:
            s += self.bracket
        for i in self.expr:
            if i is not None:
                s += str(i) + symbols.SPACE
            if self.operator == symbols.PLUS:
                s += symbols.PLUS + symbols.SPACE
        if self.operator == symbols.PLUS:
            s = s[:-3]
        else:
            s = s[:-1]
        if self.bracket != symbols.NONE:
            s += self.endBracket
        return s

    def printVal(self):
        s = ''
        if self.bracket != symbols.NONE:
            s += self.bracket
        for i in self.expr:
            if i is not None:
                s += i.printVal() + symbols.SPACE
            if self.operator == symbols.PLUS:
                s += symbols.PLUS + symbols.SPACE
        if self.operator == symbols.PLUS:
            s = s[:-3]
        else:
            s = s[:-1]
        if self.bracket != symbols.NONE:
            s += self.endBracket
        return 'tupExpr'+s
        
        
    def map(self, params, sort, onto):
        # if not tuple --> False
        if onto.type != 'description-tuple':
            return False
        # for entity in expression list
        for i in range(len(self.expr)):
            entity = self.expr[i]
            # if entity is a parameter with kleene star e.g. H*
            if entity.type == 'description-parameter' and entity.hasQuantifier():
                # if entity is the last expression
                if i == len(self.expr) -1:
                    tpl =  kleeneTuple(self.bracket, symbols.COMMA)
                    # add the rest of the entities not yet mapped in onto into a tuple
                    # e.g. mapping [h H*] onto [h, A, B, C, D]
                    #       --> h  mapped to h
                    #           H* mapped to (A, B, C, D)
                    for j in range(i,len(onto)):
                        tpl.addEntity(onto.getEntity(j))
                    # map the parameter into a tuple
                    entity.map(params, sort, tpl)
                    return True
                else:
                    return False
            elif not entity.map(params, sort, onto.getEntity(i)):
                return False
        return True

    def evaluate(self, params, sort):
        if self.operator == symbols.PLUS:
            result = self.left()
            if result is not None:
                result = result.evaluate(params, sort)
            else:
                result = params.retrieve(sort, symbols.VALUE)
            result.sum(self.right().evaluate(params, sort))
            return result
        if self.bracket == symbols.NONE:
            result = tupleDesc(symbols.PAREN_OPEN, symbols.COMMA)
        else:
            result = tupleDesc(self.bracket, symbols.COMMA)
        for i in range(len(self.expr)):
            ex = self.expr[i]
            value = ex.evaluate(params, sort)
            if ex.type == 'description-parameter' and value.type == 'description-kleeneTuple':
                for i in value.tuple:
                    result.addEntity(i)
            elif value.type == 'description-literal' and (value.name == 'e' or value.name == 'nil'):
                if i == len(self.expr)-1 and len(result.tuple) == 0:
                    result.addEntity(ex)
                else:
                    continue
            else:
                if ex.type == 'description-parameter' and ex.quantifier == 0 and value.type == 'description-tuple' and value.bracket != '<':
                    result.bracket = value.bracket
                    result.endBracket = value.endBracket
                    result.separator = value.separator

                    for i in value.tuple:
                        result.addEntity(i)
                else:
                    result.addEntity(value)
        #print(result)
        return result

    def contains(self, other):
        if other.isEmpty():
            return True
        if self.isEmpty():
            return False
        for i in other.expr:
            check = False
            for j in self.expr:
                if i.type == j.type and j.contains(i):
                    check = True
                    break
            if not check:
                return False
        return True
        
        
    def similarFragment(self, other):
        if self.type != other.type:
            return False
        if self.bracket != other.bracket or self.operator != other.operator:
            return False
        if len(self.expr) != len(other.expr):
            return False
        for i in range(len(self.expr)):
            if not self.expr[i].similarFragment(other.expr[i]):
                return False
        return True
    
    def mergeFragment(self, other):
        for i in range(len(self.expr)):
            self.expr[i].mergeFragment(other.expr[i])
    
    @staticmethod
    def parse(desc, reader):
        bracket = reader.token()
        reader.newToken()
        entity = desc.parseEntity(desc, reader)

        if reader.token() == symbols.COMMA or reader.token() == symbols.SEMICOLON:
            result = tupleDesc(bracket, reader.token(), entity)
            result.parseContinue(desc, reader)
            return result
        elif reader.token() == symbols.ANGLE_CLOSE or reader.token() == symbols.SQUARE_CLOSE:
            if entity is None:
                result = tupleDesc(bracket, symbols.UNKNOWN)
            elif entity.type == 'description-parameter' and entity.quantifier == '*':
                result = tupleExpression._initBracket(bracket, entity)
            else:
                result = tupleDesc(bracket, symbols.UNKNOWN, entity)
            if reader.token() != result.endBracket:
                raise ParsingError(reader, "expected a '" + result.endbracket + "'")
            reader.newToken()
            return result
        result = tupleExpression._initBracket(bracket, entity)
        while reader.token() != result.endBracket:
            result.expr.append(desc.parseEntity(desc, reader))
        reader.newToken()
        return result
        
    
    @staticmethod
    def parseContinue(desc, first, second, reader):
        result = tupleExpression._initOperator(symbols.NONE)
        result.expr.append(first)
        result.expr.append(second)
        tuples = first.type == 'description-tuple'
        if tuples:
            tuples = (second.type == 'description-tuple' or second.type == 'description-literal')
        nextEntity = desc.parseEntity(desc, reader)
        while nextEntity is not None:
            if tuples:
                tuples = nextEntity.type == 'description-tuple'
            result.expr.append(nextEntity)
            nextEntity = desc.parseEntity(desc, reader)
        if not tuples:
            if first.type == 'description-parameter' or second.type == 'description-parameter':
                return result
            if ((first.type == 'description-numberLiteral' or second.type == 'description-stringLiteral')
                and second.type == 'description-tuple' and (not second.isEmpty()) and 
                first.__class__ == second.getEntity(0).__class__ and
                (len(second) == 1 or first.__class__ == second.getEntity(1).__class__)):
                    return result
        tuple = tupleDesc(symbols.NONE, symbols.NONE)
        for entity in result.expr:
            if not entity.type == 'description-literal' or entity == desc.literal('e') or entity == desc.literal('nil'):
                tuple.addEntity(entity)
        del result.expr[:]
        return tuple
