# -*- coding: utf-8 -*-
"""
Created on Mon Nov  7 10:28:01 2016

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import super
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from .individual import individual
from ..struct.coordinate import cn1
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D

class plane2D(individual):
    
    type = 'ind-plane2D'
    
    def __init__(self, sort, *args, **kwargs):
        super(plane2D, self).__init__(sort)
    
    def toString(self, printKey):
        s = ''
        if printKey and self.key is not None:
            s += '#'+str(self.key)+' '
        return s+'plane'
        
    def equals(self, other):
        return True
        
    def isNil(self):
        return False
    
    def compareValue(self, other):
        if other.type == 'ind-plane2D':
            return 0
        return False
        
    def findFloatValue(self, element):
        if element.type == 'ind-plane2D':
            return True
        return False
        
    @staticmethod
    def compareAngle(center, first, second):
        return (first.product(second) + center.product(first.subtract(second))).sign()
        
    @staticmethod
    def compareRelAngle(center, ref, first, second):
        a = plane2D.compareAngle(center, ref, first)
        b = plane2D.compareAngle(center, ref, second)
        if a == b:
            if a != 0:
                return plane2D.compareAngle(center, first, second)
            if (ref < center) == (second < center) and (ref < center) != (first < center):
                return 1
            if (ref < center) != (second < center) and (ref < center) == (first < center):
                return -1
            return 0
        if a == 0:
            if (ref < center) == (first < center):
                return -1
            else:
                return -1 * b
        if b == 0:
            if (ref < center) == (second < center):
                return 1
        return a

    @staticmethod
    def angleContainsPin(center, p1, pin, p2):
        a = plane2D.compareAngle(center, p1, pin)
        b = plane2D.compareAngle(center, p1, p2)
        if a == b:
            if a != 0:
                return plane2D.compareAngle(center, pin, p2)
            if (p1 < center) == (p2 < center) and (p1 < center) != (pin < center):
                return 1
            return 0
        if a == 0:
            if (p1 < center) == (pin < center):
                return 0
            else:
                return -1 * b
        if b == 0:
            if (p1 < center) == (p2 < center):
                return 1
        return a
        
    @staticmethod
    def findPin(center, start, end):
        angleComp = plane2D.compareAngle(center, start, end)
        if angleComp == -1:
            return start.midpoint(end)
        mid = start.midpoint(end)
        if angleComp == 1:
            pin = center.subtract(mid.subtract(center))
            return pin
        if mid == center:
            a = vector2D(1,0)
            pin = center.add(a)
            if plane2D.compareAngle(center, start, pin) == 0:
                a = vector2D(0,1)
                pin = center.add(a)
            if plane2D.compareAngle(center, start, pin) == 1:
                pin = center.subtract(a)
            return pin
        raise ValueError('Backend error: start == end when find pin')
        
    def parallel(self, other):
        if other.type == 'ind-plane2D':
            return True
        return False
        
class plane3D(individual):
    
    type = 'ind-plane3D'
    
    def __init__(self, sort, boundary, normal_n_origin = False, normal = None, origin = None, *args, **kwargs):
        if normal_n_origin is False:        
            super(plane3D, self).__init__(sort)
            p0, p1, p2 = boundary[0], boundary[1], boundary[2]
            cc = 3
            while p0.colinear(p1,p2):
                print(p0.printRhino(), p1.printRhino(), p2.printRhino())
                p2 = boundary[cc]
                cc+=1
            self.normal = ((p2.subtract(p0)).product(p1.subtract(p0))).unit()
            if self.normal < vector3D(0,0,0):
                self.negNormal = True
                self.normal = self.normal.scale(cn1)
            else:
                self.negNormal = False
            self.scalar = self.normal.scalar(boundary[0])
            self.pRoot = self.normal.scale(self.scalar)
            
        else:
            super(plane3D, self).__init__(sort)
            self.normal = normal
            if self.normal < vector3D(0,0,0):
                self.negNormal = True
                self.normal = self.normal.scale(cn1)
            else:
                self.negNormal = False
            self.scalar = self.normal.scalar(origin)
            self.pRoot = self.normal.scale(self.scalar)
            
        if self.normal.parallel(vector3D(0,0,1)):
            self.u = vector3D(1,0,0)
            self.v = vector3D(0,1,0)
        else:
            self.u = self.normal.product(vector3D(0,0,1)).unit()
            self.v = self.normal.product(self.u).unit()
            
    def toString(self, printKey):
        s = ''
        if printKey and self.key is not None:
            s += '#'+str(self.key)+' '
        return s + 'normal: '+str(self.normal)+'pRoot: '+str(self.pRoot)
        
    def printRhino(self, dX = 0, dY = 0):
        if not self.normal.x == 0:
            dir0 = self.normal.__class__(self.normal.y, -self.normal.x, 0)
            dir1 = self.normal.__class__(self.normal.z, 0, -self.normal.x)
        elif not self.normal.y == 0:
            dir0 = self.normal.__class__(self.normal.y, -self.normal.x, 0)
            dir1 = self.normal.__class__(0, self.normal.z, -self.normal.y)
        elif not self.normal.z == 0:
            dir0 = self.normal.__class__(self.normal.z, 0, -self.normal.x)
            dir1 = self.normal.__class__(0, self.normal.z, -self.normal.y)
        p = []
        p.append(dir0.scale(10).add(self.pRoot))
        p.append(dir1.scale(10).add(self.pRoot))
        p.append(dir0.scale(-10).add(self.pRoot))
        p.append(dir1.scale(-10).add(self.pRoot))
        s='_polyline '
        for j in p:
            s += str(j.x + dX) + ',' + str(j.y + dY) + ',' + str(j.z) + ' '
        s+= str(p[0].x + dX) + ',' + str(p[0].y + dY) + ',' + str(p[0].z) + '\n'+'   _SelPolyline _PlanarSrf _Delete'
        return s 
        
    def equals(self, other):
        return self.normal == other.normal and self.scalar == other.scalar
        
    def compareValue(self, other):
        c = self.normal.compareValue(other.normal)
        if c != 0:
            return c
        return self.scalar.compareValue(other.scalar)
                
    def isNil(self):
        return self.normal.isZero()
        
    def direction(self, coords):
        return self.u.scale(coords.x).add(self.v.scale(coords.y))
        
    def position(self, coords):
        return self.pRoot.add(self.direction(coords))
        
    def localize(self, point):
        diff = point.subtract(self.pRoot)
        return vector2D(self.u.scalar(diff), self.v.scalar(diff))
        
    def projection(self, point):
        return self.position(self.localize(point))
        
    def containsPoint(self, point):
        if self.isNil():
            raise TypeError('Plane can not be nil')
        return self.normal.dotproduct(self.pRoot.subtract(point)) == 0
                
    def parallel(self, other):
        if self.isNil() or other.isNil():
            raise TypeError('Neither plane may be nil')
        return self.normal == other.normal
    
    def perpendicular(self, other):
        if self.isNil() or other.isNil():
            raise TypeError('Neither plane may be nil')
        return self.normal.perpendicular(other.normal)
            
    def parallelLine(self, line):
        if self.isNil() or line.isNil():
            raise TypeError('Neither plane nor line may be nil')
        return self.normal.dotproduct(line.direction) == 0
    
    def containsLine(self, line):
        if self.isNil() or line.isNil():
            raise TypeError('Neither plane nor line may be nil')
        return self.normal.dotproduct(line.direction) == 0 and (self.pRoot == line.root 
        or line.root.subtract(self.pRoot).dotproduct(self.normal) == 0)
    
    def perpendicularLine(self, line):
        if self.isNil() or line.isNil():
            raise TypeError('Neither plane nor line may be nil')
        return self.normal == line.unitDir
            
    def findFloatValue(self, element):
        if element.type == 'ind-plane3D':
            return False
        if (self.normal.convertFloat() == element.normal and
            self.scale.convertFloat() == element.scale):
            return self
        if self.hasAttribute():
            return self.attribute.findFloatValue(element)
        return False
        
    def replicateInd(self):
        result = plane3D(self.sort, [], True, self.normal, self.pRoot)
        return result
        
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            newPlane = self.duplicateInd()
        else:
            position0 = self.pRoot.transform(affine);
            position1 = self.pRoot.add(self.u).transform(affine)      
            position2 = self.pRoot.add(self.v).transform(affine)      
            newPlane =  plane3D(self.sort, [position0,position1,position2])
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is not None:
            count = 0
            bounds = []
            for i in transf.vects:
                if self.containsPoint(i[0]):
                    if len(bounds) == 2 and i[1].colinear(bounds[0],bounds[1]):
                        continue
                    bounds.append(i[1])
                    count += 1
                if count >= 3:
                    break
            newPlane = plane3D(self.sort, bounds)
        return newPlane
        
    def intersection(self, other):
        if self.isNil() or other.isNil():
            raise TypeError('Neither plane may be nil')
        if self.parallel(other):
            raise ValueError('The two planes are parallel to one another')
        prod = self.normal.product(other.normal)
        
        pDir = prod.product(self.normal)
        d = (other.pRoot.subtract(self.pRoot).dotproduct(other.normal)) / (pDir.dotproduct(other.normal))
        return prod, pDir.scale(d).add(self.pRoot)

    def intersectionLine(self, line):
        if self.isNil() or line.isNil():
            raise TypeError('Neither plane nor line may be nil')
        if self.parallelLine(line):
            raise ValueError('The two object are parallel to one another')
        d = (self.pRoot.subtract(line.root).dotproduct(self.normal)) / (line.direction.dotproduct(self.normal))
        return line.direction.scale(d).add(line.root)
