# -*- coding: utf-8 -*-
"""
Created on Thu Nov 24 09:24:16 2016

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
"""
from future import standard_library
standard_library.install_aliases()
"""
from builtins import object
from enum import Enum

from ...io.parseReader import parseReader
from ...struct.coordinate import coordinate as co

class colorMode(object):
    class Model(Enum):
        RGB = 0
        HSV = 1
        
    class Behavior(Enum):
        RGB_MAX = 0
        RGB_ALPHA = 1
        RGB_AVG = 2
        RGB_SUM = 3
    
    MAX_ALPHA = co(1) #transparency of color
    MAX_RGB_VALUES = [co(255)]
    MAX_HSV_VALUES = [co(1),co(1),co(255)]
        
        
    """
    input: argList (string)
               color model             color behavior          value1/alpha   value2/value1   value3/value2   *empty*/value3
    argList = 'HSV/HSB/RGB  RGB_MAX/RGB_ALPHA/RGB_AVG/RGB_SUM  float1         float2          float3          float4(optional),'
    """
    def __init__(self, argList):
        m = colorMode.Model
        b = colorMode.Behavior
        
        reader = parseReader(argList)
        if reader.token() != parseReader.IDENTIFIER:
            raise ValueError(reader, "Expected an identifier argument ('a')")
        modelStr = reader.value()
        if modelStr == 'HSV' or modelStr == 'HSB':
            model = m.HSV
        elif modelStr == 'RGB':
            model = m.RGB
        else:
            raise ValueError(reader, "Expected 'RGB', 'HSV' or 'HSB'")
        behavior = b.RGB_MAX
        if reader.newToken() == ',':
            self.model = model
            self.behavior = behavior
            self.max_alpha = 0
            self.max_cvalues = None
            return
        
        if reader.token() != "a":
            raise ValueError(reader, "Expected an identifier argument ('a')")
        behaviorstr = reader.value()
        if behaviorstr == 'RGB_ALPHA' or behaviorstr == 'ALPHA':
            behavior = b.RGB_ALPHA
        elif behaviorstr == 'RGB_AVG' or behaviorstr == 'AVG':
            behavior = b.RGB_AVG
        elif behaviorstr == 'RGB_SUM' or behaviorstr == 'SUM':
            behavior = b.RGB_SUM
        elif behaviorstr == 'RGB_MAX' or behaviorstr == 'MAX':
            behavior = b.RGB_MAX
        else: 
            raise ValueError(reader, "Expected 'RGB_MAX', 'RGB_ALPHA', 'RGB_AVG' or 'RGB_SUM'")
                    
        if reader.newToken() == ',':
            self.model = model
            self.behavior = behavior
            self.max_alpha = 0
            self.max_cvalues = None
            return

        if reader.token() != "0":
            raise ValueError(reader, "Number expected ('0')")
        alpha = 0;
        value = reader.value()

        if reader.previewToken() != ',' :
            values = [value,None,None,None]
            i = 1
            while i < 4 and reader.newToken() != ',':
                if reader.token() != "0":
                    raise ValueError(reader, "Number expected ('0')")
                values[i] = reader.value()
                i+=1
            
            if i == 4:
                if behavior == b.RGB_ALPHA:
                    alpha = values[0]
                values = values[1:4]
        else:
            values = [value]

        for i in range(len(values)):
            if values[i] is not None:
                values[i] = co(values[i])
        self.model = model
        self.behavior = behavior
        self.max_alpha = alpha
        self.max_cvalues = values
        
    @property
    def max_cvalues(self):
        if self.__max_cvalues is not None:
            return self.__max_cvalues
        if self.model == colorMode.Model.HSV:
            return self.MAX_HSV_VALUES
        return self.MAX_RGB_VALUES
            
    @max_cvalues.setter
    def max_cvalues(self, val):
        self.__max_cvalues = val

    
    @property
    def max_alpha(self):
        if self.__max_alpha > 0:
            return self.__max_alpha
        return self.MAX_ALPHA
    

    @max_alpha.setter
    def max_alpha(self, val):
        self.__max_alpha = co(val)

    